# routine to fit a Chebyshev polynomial to a set of absorption correction
# values for use in program gformat.
#
# $Revision: 1.1 $    $Date: 2004/06/15 19:08:07 $
#
# evaluate the Chebyshev polynomial with coefficients A at point x
# coordinates are rescaled from $xmin=-1 to $xmax=1
proc chebeval {A x xmin xmax} {
    set xs [expr {-1 + 2 * (1.*$x - $xmin) / (1.*$xmax - 1.*$xmin)}]
    set Tpp 0
    set Tp 0
    set total 0
    foreach a $A {
	if {$Tpp == $Tp && $Tp == 0} {
	    set T 1
	} elseif {$Tpp == 0} {
	    set T $xs
	} else {	
	    set T [expr {2. * $xs * $Tp - $Tpp}]
	}
	set total [expr {$total + $a * $T}]
	set Tpp $Tp
	set Tp $T
    }
    return $total
}

# perform a Gauss-Newton fit to optimize Chebyshev coefficients A
proc chebGN {X Y S A xmin xmax "damp 0.75"} {
    # Gauss-Newton
    if {[llength $X] != [llength $Y]} return
    set xnorm [expr {2. / ($xmax - $xmin)}]
    # denominator
    set sum2 0.
    foreach x $X s $S {
	set xs [expr {-1 + (1.*$x - $xmin) * $xnorm}]
	set Tpp 0
	set Tp 0
	foreach a1 $A {
	    if {$Tpp == $Tp && $Tp == 0} {
		set T 1
	    } elseif {$Tpp == 0} {
		set T $xs
	    } else {	
		set T [expr {2. * $xs * $Tp - $Tpp}]
	    }
	    set sum2 [expr {$sum2 + $T /($s*$s)}]
	    set Tpp $Tp
	    set Tp $T
	}
    }
    # Evaluate Ycalc & sum(delta2)
    set sumd2 0.
    foreach x $X y $Y {
#	set xs [expr {-1 + (1.*$x - $xmin) * $xnorm}]
	lappend Ycalc [set yc [chebeval $A $x $xmin $xmax]]
	set sumd2 [expr {$sumd2 + ($y - $yc)*($y - $yc)}]
    }
    set k -1
    foreach a $A {incr k; set sum($k) 0}
    foreach x $X y $Y yc $Ycalc s $S {
	set xs [expr {-1 + (1.*$x - $xmin) * $xnorm}]
	set Tpp 0
	set Tp 0
	set k -1
	foreach a $A {
	    incr k
	    if {$Tpp == $Tp && $Tp == 0} {
		set T 1
	    } elseif {$Tpp == 0} {
		set T $xs
	    } else {	
		set T [expr {2. * $xs * $Tp - $Tpp}]
	    }
	    set sum($k) [expr {$sum($k) + ($T * ($yc - $y))/($s*$s)}]
	    set Tpp $Tp
	    set Tp $T
	}
    }
    set sumd2r $sumd2
    set d $damp
    # compute new cheb terms
    while {$d > $damp/32} {
	set k -1
	set Anew {}
	foreach a $A {
	    incr k
	    lappend Anew [expr {$a - $d*($sum($k) / $sum2)}]
	}
	# Evaluate new Ycalc & sum(delta2)
	set sumd2r 0.
	foreach x $X y $Y {
#	    set xs [expr {-1 + (1.*$x - $xmin) * $xnorm}]
	    set yc [chebeval $Anew $x $xmin $xmax]
	    set sumd2r [expr {$sumd2r + ($y - $yc)*($y - $yc)}]
	}
	# are these shifts an improvement?
	if {$sumd2r < $sumd2} {
	    # are we converged?
	    if {[expr {($sumd2-$sumd2r)/$sumd2}] < 0.0001} {return ""}
	    return $Anew
	}
	set d [expr {$d/2.}]
    }
    return ""
}


# evaluate the best-fit Chebyschev polynomial to a set of X and Y values.
# order is the # of terms 
# o is order
proc FitChebFunc {X Y o Xmin Xmax} {
    # zero the matrix and vector
    for {set j 0} {$j < $o} {incr j} {
        set sum($j) 0.
        for {set i 0} {$i <= $j} {incr i} {
            set sum(${i}_$j) 0.
        }
    }
    #global octave
    #set octave {}
    #append octave {des = [}
    set Xnorm  [expr {2. / (1.*$Xmax - 1.*$Xmin)}]
    foreach y $Y xt $X {
        # rescale x
        set x [expr {-1 + $Xnorm * (1.*$xt - $Xmin)}]
        # compute derivatives at point x
        set derivlist {}
        set Tpp 0
        set Tp 0
        # compute the Chebyschev term Tn(xs)
        for {set i 0} {$i < $o} {incr i} {
            if {$Tpp == $Tp && $Tp == 0} {
                set T 1
            } elseif {$Tpp == 0} {
                set T $x
            } else {
                set T [expr {2. * $x * $Tp - $Tpp}]
            }
            lappend derivlist $T
            set Tpp $Tp
            set Tp $T
        }
        #append octave " $derivlist ;\n"
        # compute matrix elements
        for {set j 0} {$j < $o} {incr j} {
            set Tj [lindex $derivlist $j]
            # weighted
            # set sum($j) [expr {$sum($j) + $y * $Tj / ($sigma*$sigma)}]
            set sum($j) [expr {$sum($j) + $y * $Tj}]
            for {set i 0} {$i <= $j} {incr i} {
                set Ti [lindex $derivlist $i]
                # weighted
                # set sum(${i}_$j) [expr {$sum(${i}_$j) + $Ti * $Tj / ($sigma * $sigma)}]
                set sum(${i}_$j) [expr {$sum(${i}_$j) + $Ti * $Tj}]
            }
        }
    }
    # populate the matrix & vector in La format
    lappend V 2 $o 0
    lappend A 2 $o $o
    for {set i 0} {$i < $o} {incr i} {
        lappend V $sum($i)
        for {set j 0} {$j < $o} {incr j} {
            if {$j < $i} {
                lappend A $sum(${j}_$i)
            } else {
                lappend A $sum(${i}_$j)
            }
        }
    }
    set termlist {}
    if {[catch {
        set termlist [lrange [La::msolve $A $V] 3 end]
    }]} {
        tk_dialog .singlar "Singular Matrix" \
            "Unable to fit function: singular matrix. Too many terms or something else is wrong." ""\
            0 OK
    }
    return $termlist
}


proc Compute {} {
    global A MuDotR argv0 Normalize Order
    set xmin 0
    set xmax 90
    set Xlist {}
    set Ylist {}
    for {set j 1} {$j <= 19} {incr j 1} {
	lappend Xlist [.a.x$j get]
	lappend Ylist [.a.y$j get]
    }
    catch {.g.g element create data -label input -color red -symbol plus  -line 0}
    catch {.g.g element config data -xdata $Xlist -ydata $Ylist}
    set A [FitChebFunc $Xlist $Ylist $Order $xmin $xmax]
    set xlist {}
    set ylist {}
    for {set j 0} {$j <= 90} {incr j 1} {
	lappend xlist $j
	lappend ylist [chebeval $A $j $xmin $xmax]
    }
    catch {.g.g element create fit -label fit -color black -symbol "" -line 2}
    catch {.g.g element config fit -xdata $xlist -ydata $ylist}
    set sumd 0
    for {set j 1} {$j <= 19} {incr j 1} {
	set x [.a.x$j get]
	set y [format %.4f [chebeval $A $x $xmin $xmax] ]
	.a.f$j config -text $y
	set d [expr {[.a.y$j get] - $y}]
	set sumd [expr {$sumd + $d*$d}]
	.a.d$j config -text [format %.5f $d]
    }
    #puts "A = $A"
    .a.dend config -text [format %.5f [expr {sqrt($sumd)/19}]]

    set norm 1
    if {$Normalize} {set norm [lindex $Ylist 0]}
    set fp [open abs.corr.$MuDotR w]
    puts $fp "Mu.R = $MuDotR, fit by $argv0"
    puts -nonewline $fp [llength $A],
    foreach a $A {
	puts -nonewline $fp [format %.6f, [expr $a/ $norm ]]
    }
    puts $fp ""
    close $fp
    tk_dialog .written "Results Written" \
	"Results saved in file abs.corr.$MuDotR" "" 0 OK

}

proc LoadFile {file} {
    if {$file == ""} {bell;return}
    if {![file exists $file]} {bell;return}
    set fp [open $file r]
    # first line is Mu.R
    set ::MuDotR [string trim [gets $fp]]
    # rest of file is 19 numbers, A* for 0-90 degrees theta
    set txt [string trim [read $fp]]
    regsub -all "," $txt " " txt
    close $fp
    set i -1
    for {set j 1} {$j <= 19} {incr j 1} {
	incr i 
	set ::Y($j) [lindex $txt $i]
    } 
}

set MuDotR 1.8
set Ylist {
    16.3  16.0  15.4  14.4  13.3  12.2  11.1  10.2  9.32  8.59  
    7.97  7.45  7.01  6.63  6.33  6.09  5.92  5.81  5.78
}

set i -5
set Xlist {}
set Slist {}

label .c0 -text "m" -font symbol
label .c1 -text ". R:"
entry .c2 -width 8 -textvariable MuDotR
checkbutton .c3 -text "Normalize Chebyschev Coefficients" -variable Normalize
grid .c0 .c1 .c2 .c3 -sticky w
grid [frame .d ] -columnspan 99
label .d.1 -text "Chebyschev order"
tk_optionMenu .d.2 Order 2 4 6 8 10 12 14 
set Order 6
grid .d.1 .d.2 -sticky w
grid [frame .a ] -columnspan 99
grid [frame .b]  -columnspan 99
button .b.0 -text Load -command {
    LoadFile [tk_getOpenFile -parent . -filetypes {
	{"data files" *.dat} {"all files" *.*}
    }]
}
button .b.1 -text Fit -command Compute
button .b.2 -text Quit -command exit
grid .b.0 .b.1 .b.2
label .a.l -text #
label .a.x -text Theta
label .a.y -text A*
label .a.f -text Fit
label .a.d -text difference
grid .a.l .a.x .a.y .a.f .a.d

for {set j 1} {$j <= 19} {incr j 1} {
    label .a.l$j -text $j
    entry .a.x$j -width 8 -textvariable X($j)
    entry .a.y$j -width 8 -textvariable Y($j)
    label .a.f$j -width 10
    label .a.d$j -width 10
    grid .a.l$j .a.x$j .a.y$j .a.f$j .a.d$j 
}
label .a.fend -text rms
label .a.dend
grid x x x .a.fend .a.dend

set j 0
foreach y $Ylist {
    incr j
    set X($j) [expr {5.*($j-1)}]
    set Y($j) $y
}

lappend auto_path [file dirname [info script]]
package require La
catch {package require BLT}
toplevel .g
grid rowconfig .g 0 -weight 1
grid columnconfig .g 0 -weight 1
wm withdraw .g
catch {
    blt::graph .g.g
    grid .g.g -sticky news -col 0 -row 0
    wm deiconify .g
}
